/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>


#define D_ARG_OUT_FILE "--out-file="
#define D_ARG_IN_FILE  "--in-file="
#define D_ARG_LINE_LEN "--line-len="
#define D_ARG_NO_NAME  "--no-name"

#define D_LINE_LEN 1024


typedef enum
{
  E_FALSE = 0,
  E_TRUE = 1
} TE_BOOL;


#define D_COLUMNS 5
typedef struct
{
  unsigned char auc_column[ D_COLUMNS ];
} TS_CHAR;


#define D_FIRST_CHAR 33
#define D_LAST_CHAR 127
const TS_CHAR ats_char[ 95 ] =
{
  {  0,  0, 79,  0,  0}, // ! (33)
  {  0,  7,  0,  7,  0}, // "
  { 20,127, 20,127, 20}, // #
  { 36, 42,127, 42, 18}, // $
  { 35, 19,  8,100, 98}, // %
  { 54, 73, 85, 34, 80}, // &
  {  0,  5,  3,  0,  0}, // '
  {  0, 28, 34, 65,  0}, // (
  {  0, 65, 34, 28,  0}, // )
  { 20,  8, 62,  8, 20}, // *
  {  8,  8, 62,  8,  8}, // +

//{  0, 80, 48,  0,  0}, // ,
  {  0,160, 96,  0,  0}, // ,

  {  8,  8,  8,  8,  8}, // -
  {  0, 96, 96,  0,  0}, // .
  { 32, 16,  8,  4,  2}, // /
  { 62, 81, 73, 69, 62}, // 0
  {  0, 66,127, 64,  0}, // 1
  { 66, 97, 81, 73, 70}, // 2
  { 33, 65, 69, 75, 49}, // 3
  { 24, 20, 18,127, 16}, // 4
  { 39, 69, 69, 69, 57}, // 5
  { 60, 74, 73, 73, 48}, // 6
  {  1,113,  9,  5,  3}, // 7
  { 54, 73, 73, 73, 54}, // 8
  {  6, 73, 73, 41, 30}, // 9
  {  0, 54, 54,  0,  0}, // :

//{  0, 86, 54,  0,  0}, // ;
  {  0,182,118,  0,  0}, // ;

  {  8, 20, 34, 65,  0}, // <
  { 20, 20, 20, 20, 20}, // =
  {  0, 65, 34, 20,  8}, // >
  {  2,  1, 81,  9,  6}, // ?
  { 50, 73,121, 65, 62}, // @
  {126, 17, 17, 17,126}, // A
  {127, 73, 73, 73, 54}, // B
  { 62, 65, 65, 65, 34}, // C
  {127, 65, 65, 34, 28}, // D
  {127, 73, 73, 73, 65}, // E
  {127,  9,  9,  9,  1}, // F
  { 62, 65, 73, 73,122}, // G
  {127,  8,  8,  8,127}, // H
  {  0, 65,127, 65,  0}, // I
  { 32, 64, 65, 63,  1}, // J
  {127,  8, 20, 34, 65}, // K
  {127, 64, 64, 64, 64}, // L
  {127,  2, 12,  2,127}, // M
  {127,  4,  8, 16,127}, // N
  { 62, 65, 65, 65, 62}, // O
  {127,  9,  9,  9,  6}, // P
  { 62, 65, 81, 33, 94}, // Q
  {127,  9, 25, 41, 70}, // R
  { 70, 73, 73, 73, 49}, // S
  {  1,  1,127,  1,  1}, // T
  { 63, 64, 64, 64, 63}, // U
  { 31, 32, 64, 32, 31}, // V
  { 63, 64, 48, 64, 63}, // W
  { 99, 20,  8, 20, 99}, // X
  {  7,  8,112,  8,  7}, // Y
  { 97, 81, 73, 69, 67}, // Z
  {  0,127, 65, 65,  0}, // [
  {  2,  4,  8, 16, 32}, // '\'
  {  0, 65, 65,127,  0}, // ]
  {  4,  2,  1,  2,  4}, // ^

//{ 64, 64, 64, 64, 64}, // _
  {128,128,128,128,128}, // _

  {  0,  1,  2,  4,  0}, // `
  { 32, 84, 84, 84,120}, // a
  {127, 72, 68, 68, 56}, // b
  { 56, 68, 68, 68, 32}, // c
  { 56, 68, 68, 72,127}, // d
  { 56, 84, 84, 84, 24}, // e
  {  8,126,  9,  1,  2}, // f

//{ 12, 82, 82, 82, 62}, // g
  { 24,164,164,164,124}, // g

  {127,  8,  4,  4,120}, // h
  {  0, 68,125, 64,  0}, // i

//{ 32, 64, 68, 61,  0}, // j
  { 64,128,132,125,  0}, // j

  {127, 16, 40, 68,  0}, // k
  {  0, 65,127, 64,  0}, // l
  {124,  4, 24,  4,120}, // m
  {124,  8,  4,  4,120}, // n
  { 56, 68, 68, 68, 56}, // o

//{124, 20, 20, 20,  8}, // p
  {252, 36, 68, 68, 56}, // p
//{252, 36, 36, 36, 24}, // p

//{  8, 20, 20, 24,124}, // q
  { 56, 68, 68, 36,252}, // q
//{ 24, 36, 36, 36,252}, // q

  {124,  8,  4,  4,  8}, // r
  { 72, 84, 84, 84, 32}, // s
  {  4, 63, 68, 64, 32}, // t
  { 60, 64, 64, 32,124}, // u
  { 28, 32, 64, 32, 28}, // v
  { 60, 64, 48, 64, 60}, // w
  { 68, 40, 16, 40, 68}, // x

//{ 12, 80, 80, 80, 60}, // y
  { 28,160,160,160,124}, // y

  { 68,100, 84, 76, 68}, // z
  {  0,  8, 54, 65,  0}, // {
  {  0,  0,127,  0,  0}, // |
  {  0, 65, 54,  8,  0}, // }
  { 16,  8, 24, 16,  8}  // ~
}; // TS_CHAR ats_char[ 95 ]



void r_help( void );

void r_get_line( unsigned char * const puc_dest,
                 const size_t o_dest_size,
                 FILE * const po_source );

FILE *r_open_output( const char * const pc_file_name );

void r_big_text( FILE * const po_out_file,
                 const unsigned char * puc_text,
                 const int i_max_line_len,
                 const TE_BOOL b_no_name );



int main( int argc, char *argv[] )
{
  unsigned char auc_line[ D_LINE_LEN + 1 ];

  FILE *po_out_file;
  FILE *po_in_file;

  char *pc_arg_out_file;
  char *pc_arg_in_file;
  int i_arg_line_len;
  TE_BOOL b_arg_no_name;

  int i_count;

  // Process all program arguments
  pc_arg_out_file = NULL;
  pc_arg_in_file = NULL;
  i_arg_line_len = 0;
  b_arg_no_name = E_FALSE;
  for ( i_count = 1; i_count < argc; i_count++ )
  {
    if ( strncmp( argv[i_count], D_ARG_OUT_FILE, strlen( D_ARG_OUT_FILE ) ) == 0 )
    {
      pc_arg_out_file = argv[i_count] + strlen( D_ARG_OUT_FILE );
    }
    else if ( strncmp( argv[i_count], D_ARG_IN_FILE, strlen( D_ARG_IN_FILE ) ) == 0 )
    {
      pc_arg_in_file = argv[i_count] + strlen( D_ARG_IN_FILE );
    }
    else if ( strncmp( argv[i_count], D_ARG_LINE_LEN, strlen( D_ARG_LINE_LEN ) ) == 0 )
    {
      i_arg_line_len = atoi( argv[i_count] + strlen( D_ARG_LINE_LEN ) );
    }
    else if ( strcmp( argv[i_count], D_ARG_NO_NAME ) == 0 )
    {
      b_arg_no_name = E_TRUE;
    }
    else
    {
      fprintf( stderr, "Invalid argument specified: '%s'\n", argv[i_count] );
      r_help();
      exit( 1 );
    }
  } // for ( i_count = 1; i_count < argc; i_count++ )

  // Check for the mantadory argument
  if ( pc_arg_out_file == NULL )
  {
    fprintf( stderr, "Error, no output file specified\n" );
    r_help();
    exit( 1 );
  }


  // Interactive or batch mode?
  if ( pc_arg_in_file == NULL )
  {
    // Interactive mode

    // Get first line
    r_get_line( auc_line, D_LINE_LEN + 1, stdin );
    while ( auc_line[0] != '\0' )
    {
      // Open output file, deleting old contents
      po_out_file = r_open_output( pc_arg_out_file );

      // Create 4 line big text
      r_big_text( po_out_file, auc_line, i_arg_line_len, b_arg_no_name );

      // Close output file
      fclose( po_out_file );

      // Get another line
      r_get_line( auc_line, D_LINE_LEN + 1, stdin );
    } // while ( auc_line[0] != '\0' )
  }
  else // if ( pc_arg_in_file == NULL )
  {
    // Batch mode

    // Open input file
    po_in_file = fopen( pc_arg_in_file, "rt" );
    if ( po_in_file == NULL )
    {
      fprintf( stderr, "Error, input file not found\n" );
      exit( 1 );
    }

    // Open output file, deleting old contents
    po_out_file = r_open_output( pc_arg_out_file );

    // Get first line
    r_get_line( auc_line, D_LINE_LEN + 1, po_in_file );
    while ( auc_line[0] != '\0' )
    {
      // Create 4 line big text
      r_big_text( po_out_file, auc_line, i_arg_line_len, b_arg_no_name );

      // Get another line
      r_get_line( auc_line, D_LINE_LEN + 1, po_in_file );
    } // while ( auc_line[0] != '\0' )

    // Close output file
    fclose( po_out_file );

    // Close input file
    fclose( po_in_file );
  } // if ( pc_arg_in_file == NULL ) else

  return 0;
} // int main( int argc, char *argv[] )


void r_help( void )
{
  fprintf( stdout, "Usage:\n" );
  fprintf( stdout, "  big_text.exe " D_ARG_OUT_FILE "<outfile> [options]\n" );
  fprintf( stdout, "\n" );
  fprintf( stdout, "Program arguments:\n" );
  fprintf( stdout, "  " D_ARG_OUT_FILE  "<outfile>         Output file (mandatory)\n" );
  fprintf( stdout, "  " D_ARG_IN_FILE  "<infile>           Input file (only for batch mode)\n" );
  fprintf( stdout, "  " D_ARG_LINE_LEN  "<max line length> Display warning when output too long\n" );
  fprintf( stdout, "  " D_ARG_NO_NAME "                    Don't show player name (namehack)\n" );
} // r_help


void r_get_line( unsigned char * const puc_dest,
                 const size_t o_dest_size,
                 FILE * const po_source )
{
  size_t o_strlen;
  char *pc_result;

  // Get one line from input
  pc_result = fgets( (char *)puc_dest, o_dest_size, po_source );

  if ( pc_result != NULL )
  {
    // Remove the optional '\n' at the end of the string
    o_strlen = strlen( puc_dest );
    if ( o_strlen > 0 )
    {
      if ( puc_dest[ o_strlen - 1 ] == '\n' )
      {
        puc_dest[ o_strlen - 1 ] = '\0';
      }
    } // if ( o_strlen > 0 )
  }
  else
  {
    // EOF or an error
    *puc_dest = '\0';
  }
} // r_get_line


FILE *r_open_output( const char * const pc_file_name )
{
  FILE *po_out_file;

  // Read from file (works for colors too)
  po_out_file = fopen( pc_file_name, "wt" );
  if ( po_out_file == NULL )
  {
    fprintf( stderr, "Error, output file open error\n" );
    exit( 1 );
  }

  // Write banner
  fprintf( po_out_file, "// BIG 4 line text version 1.0\n" );
  fprintf( po_out_file, "\n" );

  return po_out_file;
} // r_open_output


void r_big_text( FILE * const po_out_file,
                 const unsigned char * puc_text,
                 const int i_max_line_len,
                 const TE_BOOL b_no_name )
{
  char aauc_lines[ 4 ][ D_LINE_LEN * ( D_COLUMNS + 1 ) + 1 ]; // D_COLUMNS + 1, because of character spacing
  unsigned char uc_char;
  int i_output_row;
  int i_column;
  int i_first_column;
  int i_last_column;
  int i_arr_index;

  int i_write_pos;
  int i_line_len;

  const unsigned char * puc_ptr;


  i_write_pos = 0;
  i_line_len = 0;

  puc_ptr = puc_text;
  while ( *puc_ptr != '\0' )
  {
    // Color?
    if ( *puc_ptr < 32 )
    {
      aauc_lines[0][i_write_pos] = *puc_ptr;
      aauc_lines[1][i_write_pos] = *puc_ptr;
      aauc_lines[2][i_write_pos] = *puc_ptr;
      aauc_lines[3][i_write_pos] = *puc_ptr;
      i_write_pos++;
    }
    else // if ( *puc_ptr < 32 )
    {
      // Space?
      if ( *puc_ptr == ' ' )
      {
        aauc_lines[0][i_write_pos] = ' ';
        aauc_lines[1][i_write_pos] = ' ';
        aauc_lines[2][i_write_pos] = ' ';
        aauc_lines[3][i_write_pos] = ' ';
        i_write_pos++;
        aauc_lines[0][i_write_pos] = ' ';
        aauc_lines[1][i_write_pos] = ' ';
        aauc_lines[2][i_write_pos] = ' ';
        aauc_lines[3][i_write_pos] = ' ';
        i_write_pos++;

        i_line_len += 2;
      }
      else // if ( *puc_ptr == ' ' )
      {
        // Other char?
        if (
             ( *puc_ptr >= D_FIRST_CHAR ) &&
             ( *puc_ptr <= D_LAST_CHAR )
           )
        {
          i_arr_index = (*puc_ptr) - D_FIRST_CHAR;
          // Ignore blank columns at the start
          i_first_column = 0;
          while (
                  ( i_first_column < D_COLUMNS ) &&
                  ( ats_char[ i_arr_index ].auc_column[ i_first_column ] == 0 )
                )
          {
            i_first_column++;
          }
          // Ignore blank columns at the end
          i_last_column = D_COLUMNS - 1;
          while (
                  ( i_last_column >= 0 ) &&
                  ( ats_char[ i_arr_index ].auc_column[ i_last_column ] == 0 )
                )
          {
            i_last_column--;
          }
          for ( i_column = i_first_column; i_column <= i_last_column; i_column++ )
          {
            // Create the 4 lines of text
            for ( i_output_row = 0; i_output_row < 4; i_output_row++ )
            {
              uc_char = ats_char[ i_arr_index ].auc_column[ i_column ];
              uc_char = ( uc_char >> ( i_output_row * 2 ) ) & 3;
              switch ( uc_char )
              {
                case 0:
                  aauc_lines[i_output_row][i_write_pos] = ' ';
                  break;
                case 1:
                  aauc_lines[i_output_row][i_write_pos] = (unsigned char)150;
                  break;
                case 2:
                  aauc_lines[i_output_row][i_write_pos] = (unsigned char)139;
                  break;
                case 3:
                  aauc_lines[i_output_row][i_write_pos] = (unsigned char)37;
                  break;
              } // switch ( uc_char )
            } // for ( i_output_row = 0; i_output_row < 4; i_output_row++ )
            i_write_pos++;
            i_line_len++;
          } // for ( i_column = i_first_column; i_column <= i_last_column; i_column++ )
        } // if ( ... )
      } // if ( *puc_ptr == ' ' ) else
      // Add character spacing (Note: Change aauc_lines size if changing this)
      if ( *(puc_ptr + 1) != '\0' )
      {
        aauc_lines[0][i_write_pos] = ' ';
        aauc_lines[1][i_write_pos] = ' ';
        aauc_lines[2][i_write_pos] = ' ';
        aauc_lines[3][i_write_pos] = ' ';
        i_write_pos++;
        i_line_len++;
      } // if ( *(puc_ptr + 1) != '\0' )
    } // if ( *puc_ptr < 32 ) else
    puc_ptr++;
  } // while ( *puc_ptr != '\0' )
  
  aauc_lines[0][i_write_pos] = '\0';
  aauc_lines[1][i_write_pos] = '\0';
  aauc_lines[2][i_write_pos] = '\0';
  aauc_lines[3][i_write_pos] = '\0';


  // Check line length
  if (
       ( i_max_line_len > 0 ) &&
       ( i_line_len > i_max_line_len )
     )
  {
    fprintf( stderr, "Warning: Line longer as specified: %d > %d\n", i_line_len, i_max_line_len );
    fprintf( po_out_file, "// Warning: Line longer as specified: %d > %d\n", i_line_len, i_max_line_len );
  }


  // Print input string
  fprintf( po_out_file, "// %s\n", puc_text );

  // The big text
  if ( b_no_name == E_TRUE )
  {
    fprintf( po_out_file, "say \"%c%s\"\n", '\r', aauc_lines[0] );
    fprintf( po_out_file, "say \"%c%s\"\n", '\r', aauc_lines[1] );
    fprintf( po_out_file, "say \"%c%s\"\n", '\r', aauc_lines[2] );
    fprintf( po_out_file, "say \"%c%s\"\n", '\r', aauc_lines[3] );
  }
  else
  {
    fprintf( po_out_file, "say \"%s\"\n", aauc_lines[0] );
    fprintf( po_out_file, "say \"%s\"\n", aauc_lines[1] );
    fprintf( po_out_file, "say \"%s\"\n", aauc_lines[2] );
    fprintf( po_out_file, "say \"%s\"\n", aauc_lines[3] );
  }

  // Empty line
  fprintf( po_out_file, "\n" );
} // r_big_text
